import logging
from datetime import datetime, timezone

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {"region": None, "debug": False, "silent": False}
RULES = [{"required": ["threshold_days"]}]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def discover_unused_access_keys(**kwargs):
    """Discover access keys not used over a certain amount of days"""

    # Required parameters
    threshold_days = kwargs.pop("threshold_days")

    session_objects = set_session_objects(
        kwargs["session"], clients=["iam"], region=kwargs["region"]
    )

    logger.info(f"Discovering access keys not used for more than {threshold_days} days")

    access_keys = []
    users_response = session_objects["iam_client"].list_users(MaxItems=50)
    while True:
        for user in users_response["Users"]:
            try:
                keys_response = session_objects["iam_client"].list_access_keys(
                    UserName=user["UserName"]
                )
            except Exception as e:
                logger.error(f"Failed to list access keys for {user['UserName']} - {e}")
                continue
            for key in keys_response["AccessKeyMetadata"]:
                if key["Status"] == "Active":
                    try:
                        last_used = session_objects[
                            "iam_client"
                        ].get_access_key_last_used(AccessKeyId=key["AccessKeyId"])
                    except Exception as e:
                        logger.error(
                            f"Failed to get last used date for {key['AccessKeyId']} - {e}"
                        )
                        continue
                    if "AccessKeyLastUsed" in last_used:
                        age = (
                            datetime.now(timezone.utc)
                            - last_used["AccessKeyLastUsed"]["LastUsedDate"]
                        ).days
                        if age > threshold_days:
                            access_keys.append(
                                {
                                    "UserName": user["UserName"],
                                    "AccessKeyId": key["AccessKeyId"],
                                }
                            )
        if users_response["IsTruncated"]:
            users_response = session_objects["iam_client"].list_users(
                Marker=users_response["Marker"]
            )
        else:
            break

    logger.info(f"{len(access_keys)} access keys found")
    logger.debug(access_keys)

    return access_keys



import logging

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {"debug": False, "silent": False}


@validate(DEFAULTS)
@set_logger()
@authenticate()
def whoami(**kwargs):
    """Return current entity"""

    session_objects = set_session_objects(
        kwargs["session"], clients=["sts"], region=kwargs["region"]
    )

    response = session_objects["sts_client"].get_caller_identity()

    data = {"Arn": response["Arn"], "Region": kwargs["region"]}

    return data



from .config import config


def validate(kwargs, param_rules):
    """Validate input parameters."""

    # Evaluate dynamic rules
    for rule in param_rules:
        for rule_type, params in rule.items():
            rule_func = config.get(rule_type)
            if rule_func:
                rule_func(kwargs, params)
            else:
                raise ValueError(f"Invalid rule type '{rule_type}'")

    # Evaluate logging rule
    rule_func = config.get("logging")
    rule_func(kwargs)



SERVICES = {
    "backup": ["create_backups", "delete_backups"],
    "cloudtrail": [
        "discover_user_events",
        "discover_resource_events",
        "discover_events",
    ],
    "ec2": [
        "copy_snapshots",
        "create_images",
        "create_snapshots",
        "delete_images",
        "delete_instances",
        "delete_security_groups",
        "delete_snapshots",
        "delete_volumes",
        "discover_active_regions",
        "discover_default_ebs_encryption",
        "discover_images",
        "discover_instances",
        "discover_no_ssm_instances",
        "discover_snapshots",
        "discover_tags",
        "discover_unused_security_groups",
        "discover_volumes",
        "encrypt_instance_volumes",
        "encrypt_volume",
        "modify_default_ebs_encryption",
        "modify_tags",
        "modify_volumes",
        "share_snapshots",
    ],
    "general": ["get_date"],
    "iam": [
        "discover_inactive_console_users",
        "discover_inactive_users",
        "discover_no_mfa_users",
        "discover_old_access_keys",
        "discover_old_password_users",
        "discover_unused_access_keys",
        "discover_unused_roles",
        "modify_access_keys",
        "modify_users_console_access",
        "quarantine_user",
    ],
    "sts": ["create_session", "whoami"],
    "s3": ["create_objects", "delete_objects", "discover_objects"],
}



from avtomat_aws.helpers.cli.set_output import set_output
from avtomat_aws.services.iam import discover_unused_roles

ACTION_DESCRIPTION = "Discover IAM roles not used over a certain amount of days."


def add_cli_arguments(parser):
    """Argument parsing"""

    required = parser.add_argument_group("required")

    required.add_argument(
        "--threshold_days",
        type=int,
        help="Get roles not used for more than this number of days.",
        required=True,
    )


def cli(args):
    """Command-line interface function"""

    inputs = vars(args)

    try:
        result = discover_unused_roles(**inputs)
        set_output(result, inputs)
    except Exception as e:
        print(f"Action failed - {e}")
        exit(1)



from avtomat_aws.helpers.cli.set_output import set_output
from avtomat_aws.services.ec2 import discover_instances

ACTION_DESCRIPTION = "Discover instances."


def add_cli_arguments(parser):
    """Argument parsing"""

    parser.add_argument(
        "--states",
        nargs="+",
        help="Instance states, separated by space.",
        required=False,
    )
    parser.add_argument(
        "--tags",
        nargs="+",
        help="Tags for discovery in Key=Value format, separated by space.",
        required=False,
    )
    parser.add_argument(
        "--instance_ids",
        nargs="+",
        help="Instance IDs to focus on, separated by space.",
        required=False,
    )
    parser.add_argument(
        "--invert",
        action="store_true",
        help="If instance_ids is supplied, return the ones that didn't conform to the filters.",
        required=False,
    )
    parser.add_argument(
        "--os", help="Return only Windows or Linux instances.", required=False
    )
    parser.add_argument(
        "--public",
        action="store_true",
        help="Get only public instances.",
        required=False,
    )


def cli(args):
    """Command-line interface function"""

    inputs = vars(args)

    try:
        result = discover_instances(**inputs)
        set_output(result, inputs)
    except Exception as e:
        print(f"Action failed - {e}")
        exit(1)



config = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {
        "standard": {"format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"}
    },
    "handlers": {
        "console": {"class": "logging.StreamHandler", "formatter": "standard"}
    },
    "loggers": {
        "": {"handlers": ["console"], "level": "WARNING"},
        "avtomat_aws": {"level": "INFO", "handlers": ["console"], "propagate": False},
    },
}



import logging

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {"region": None, "debug": False, "silent": False}


@validate(DEFAULTS)
@set_logger()
@authenticate()
def discover_active_regions(**kwargs):
    """Discover active regions for an account"""

    session_objects = set_session_objects(
        kwargs["session"], clients=["ec2"], region=kwargs["region"]
    )

    logger.info("Discovering active regions")

    regions = [
        region["RegionName"]
        for region in session_objects["ec2_client"].describe_regions()["Regions"]
    ]

    logger.info(f"{len(regions)} active regions found.")
    logger.debug(regions)

    return regions



import argparse
import sys
from importlib import import_module, metadata

from .services import SERVICES

VERSION = metadata.version("avtomat_aws")


def display_help():
    """Displays help for the main CLI."""
    print("usage: aaws <service> <action> [parameters]\n")
    print("Avtomat AWS CLI\n")
    print("options:")
    print("    -h, --help            Show this help message and exit.")
    print("    -v, --version         Show the current version and exit.\n")
    print("services:")
    # Display services in two columns
    services_list = list(SERVICES.keys())
    max_length = (
        max(len(service) for service in services_list) + 2
    )  # Dynamically calculate space between columns
    if len(services_list) % 2 != 0:
        services_list.append("")
    for service1, service2 in zip(services_list[::2], services_list[1::2]):
        print(f"    {service1:<{max_length}} {service2}")


def display_service_help(service):
    """Displays help for a specific service including available actions."""
    print("usage: aaws <service> <action> [parameters]\n")
    print("Avtomat AWS CLI\n")
    print("options:")
    print("    -h, --help            Show this help message and exit.")
    print("    -v, --version         Show the current version and exit.\n")
    print("actions:")
    # Display actions in two columns
    actions_list = SERVICES.get(service, [])
    max_length = (
        max(len(action) for action in actions_list) + 2
    )  # Dynamically calculate space between columns
    if len(actions_list) % 2 != 0:
        actions_list.append("")
    for action1, action2 in zip(actions_list[::2], actions_list[1::2]):
        print(f"    {action1:<{max_length}} {action2}")


def display_action_help(service, action):
    """Displays help for a specific action, including action-specific arguments."""
    # Load the module for the action so that we can dynamically generate the help message through argparse
    parser = set_parser(service, action)
    # Print the help message
    parser.print_help()


def load_module(service, action):
    """Dynamically load the module for the given service and action."""
    try:
        module_name = f"avtomat_aws.cli.{service}.{action}"
        return import_module(module_name)
    except ImportError:
        sys.stderr.write(f"Error: Unable to load module for {service} {action}\n")
        sys.exit(1)


def set_parser(service, action):
    """Dynamically set the parser for the given service and action."""

    module = load_module(service, action)
    action_description = getattr(
        module, "ACTION_DESCRIPTION", "No description available"
    )
    parser = argparse.ArgumentParser(
        prog=f"aaws {service} {action}", add_help=False, description=action_description
    )

    # Global arguments
    parser.add_argument(
        "--region",
        help="Region for operation. Leave blank for session default.",
        required=False,
    )
    parser.add_argument(
        "--debug", action="store_true", help="Increase log verbosity.", required=False
    )
    parser.add_argument(
        "--silent", action="store_true", help="Decrease log verbosity.", required=False
    )
    parser.add_argument(
        "--output", help="Output format.", choices=["table"], required=False
    )

    # Use the imported module's function to add its arguments to the parser
    if hasattr(module, "add_cli_arguments"):
        module.add_cli_arguments(parser)
        parser.set_defaults(func=module.cli)
    else:
        print("No arguments defined for this action.")
        return

    return parser


def main():
    """Main function for the CLI."""

    # Version
    if len(sys.argv) == 2 and (sys.argv[1] == "--version" or sys.argv[1] == "-v"):
        print(f"aaws {VERSION}")
        sys.exit(0)

    # Not enough arguments
    if len(sys.argv) < 3:
        display_help()
        sys.exit(1)

    # Invalid service
    if sys.argv[1] not in SERVICES:
        sys.stderr.write(f"Error: Invalid service: {sys.argv[1]}\n")
        sys.exit(1)

    # Only a service with --help or -h
    if len(sys.argv) == 3 and (sys.argv[2] == "--help" or sys.argv[2] == "-h"):
        service = sys.argv[1]
        display_service_help(service)
        sys.exit(0)

    # Invalid action
    if sys.argv[2] not in SERVICES[sys.argv[1]]:
        sys.stderr.write(f"Error: Invalid action: {sys.argv[2]}\n")
        sys.exit(1)

    # A service, action, and --help or -h
    if len(sys.argv) == 4 and (sys.argv[3] == "--help" or sys.argv[3] == "-h"):
        service, action = sys.argv[1:3]
        display_action_help(service, action)
        sys.exit(0)

    # Execute
    service, action = sys.argv[1:3]
    parser = set_parser(service, action)
    args = parser.parse_args(sys.argv[3:])
    args.func(args)


if __name__ == "__main__":
    main()



from avtomat_aws.services.ec2 import modify_default_ebs_encryption

ACTION_DESCRIPTION = "Modify default EBS encryption."


def add_cli_arguments(parser):
    """Argument parsing"""

    parser.add_argument(
        "--enable",
        action="store_true",
        help="Enable default EBS encryption.",
        required=False,
    )
    parser.add_argument(
        "--disable",
        action="store_true",
        help="Disable default EBS encryption.",
        required=False,
    )
    parser.add_argument(
        "--change_kms_key",
        action="store_true",
        help="Change the KMS Key used for encryption.",
        required=False,
    )
    parser.add_argument(
        "--kms_key_id", help="KMS Key ID to use for encryption.", required=False
    )
    parser.add_argument(
        "--reset_kms_key",
        action="store_true",
        help="Reset the KMS Key used for encryption to AWS EBS default.",
        required=False,
    )


def cli(args):
    """Command-line interface function"""

    inputs = vars(args)

    try:
        modify_default_ebs_encryption(**inputs)
    except Exception as e:
        print(f"Action failed - {e}")
        exit(1)



import logging
from datetime import datetime, timezone

from botocore.exceptions import ClientError

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {"region": None, "debug": False, "silent": False}
RULES = [{"required": ["threshold_days"]}]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def discover_inactive_console_users(**kwargs):
    """Discover IAM users with last console sign-in over a certain period"""

    # Required parameters
    threshold_days = kwargs.pop("threshold_days")

    session_objects = set_session_objects(
        kwargs["session"], clients=["iam"], region=kwargs["region"]
    )

    logger.info(
        f"Discovering users with last console sign-in over {threshold_days} days"
    )

    users = []
    users_response = session_objects["iam_client"].list_users(MaxItems=50)
    while True:
        for user in users_response["Users"]:
            try:
                password_last_used = (
                    session_objects["iam_client"]
                    .get_user(UserName=user["UserName"])["User"]
                    .get("PasswordLastUsed")
                )
            except Exception as e:
                logger.error(
                    f"Failed to get last console sign-in for {user['UserName']} - {e}"
                )
                continue
            try:
                console_access = session_objects["iam_client"].get_login_profile(
                    UserName=user["UserName"]
                )
            except ClientError as e:
                if e.response["Error"]["Code"] == "NoSuchEntity":
                    continue
                else:
                    logger.error(
                        f"Failed to get console access status for {user['UserName']} - {e}"
                    )
                    continue

            if password_last_used:
                age = (datetime.now(timezone.utc) - password_last_used).days
                if age > threshold_days:
                    users.append({"UserName": user["UserName"]})
            elif console_access.get("LoginProfile"):
                users.append({"UserName": user["UserName"]})
        if users_response["IsTruncated"]:
            users_response = session_objects["iam_client"].list_users(
                Marker=users_response["Marker"]
            )
        else:
            break

    logger.info(f"{len(users)} users found")
    logger.debug(users)

    return users



import logging
from datetime import datetime, timezone

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {
    "prefix": "",
    "modified_before": None,
    "modified_after": None,
    "name_only": False,
    "region": None,
    "debug": False,
    "silent": False,
}
RULES = [
    {"required": ["bucket"]},
    {"date_yymmdd": ["modified_before", "modified_after"]},
]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def discover_objects(**kwargs):
    """Discover objects in an S3 bucket"""

    # Required parameters
    bucket = kwargs.pop("bucket")

    if kwargs.get("modified_before"):
        kwargs["modified_before"] = datetime.strptime(
            kwargs["modified_before"], "%Y/%m/%d"
        ).replace(tzinfo=timezone.utc)
    if kwargs.get("modified_after"):
        kwargs["modified_after"] = datetime.strptime(
            kwargs["modified_after"], "%Y/%m/%d"
        ).replace(tzinfo=timezone.utc)

    session_objects = set_session_objects(
        kwargs["session"], resources=["s3"], region=kwargs["region"]
    )

    if kwargs.get("prefix"):
        logger.info(
            f"Discovering objects in '{bucket}' with prefix '{kwargs['prefix']}'"
        )
    else:
        logger.info(f"Discovering objects in '{bucket}'")

    response = (
        session_objects["s3_resource"]
        .Bucket(bucket)
        .objects.filter(Prefix=kwargs["prefix"])
    )
    objects = filter_objects(response, **kwargs)

    logger.info(f"{len(objects)} objects found")
    logger.debug(objects)

    return objects


def filter_objects(response, **kwargs):
    """Filter objects based on additional criteria"""

    filtered_objects = response

    if kwargs.get("modified_before"):
        filtered_objects = [
            obj
            for obj in filtered_objects
            if obj.last_modified < kwargs["modified_before"]
        ]

    if kwargs.get("modified_after"):
        filtered_objects = [
            obj
            for obj in filtered_objects
            if obj.last_modified > kwargs["modified_after"]
        ]

    if kwargs.get("name_only"):
        filtered_objects = [obj.key.split("/")[-1] for obj in filtered_objects]
    else:
        filtered_objects = [obj.key for obj in filtered_objects]

    return filtered_objects



import logging

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {
    "enable": False,
    "disable": False,
    "change_kms_key": False,
    "kms_key_id": None,
    "reset_kms_key": False,
    "region": None,
    "debug": False,
    "silent": False,
}
RULES = [
    {"and": ["change_kms_key", "kms_key_id"]},
    {"at_most_one": ["enable", "disable"]},
    {"at_most_one": ["change_kms_key", "reset_kms_key"]},
    {"at_least_one": ["enable", "disable", "change_kms_key", "reset_kms_key"]},
]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def modify_default_ebs_encryption(**kwargs):
    """Modify the default EBS encryption setting"""

    region = kwargs["region"]

    session_objects = set_session_objects(
        kwargs.get("session"), clients=["ec2"], region=region
    )

    logger.info(f"Modifying default EBS encryption setting")

    if kwargs.get("enable"):
        session_objects["ec2_client"].enable_ebs_encryption_by_default()
        logger.info(f"{region}: Enabled default EBS encryption")
    if kwargs.get("disable"):
        session_objects["ec2_client"].disable_ebs_encryption_by_default()
        logger.info(f"{region}: Disabled default EBS encryption")
    if kwargs.get("change_kms_key"):
        session_objects["ec2_client"].modify_ebs_default_kms_key_id(
            KmsKeyId=kwargs["kms_key_id"]
        )
        logger.info(f"{region}: Changed default KMS key to {kwargs['kms_key_id']}")
    if kwargs.get("reset_kms_key"):
        session_objects["ec2_client"].reset_ebs_default_kms_key_id()
        logger.info(f"{region}: Reset default KMS key")

    return



from avtomat_aws.helpers.cli.set_output import set_output
from avtomat_aws.services.ec2 import copy_snapshots

ACTION_DESCRIPTION = "Move EC2 snapshots between regions or accounts."


def add_cli_arguments(parser):
    """Argument parsing"""

    required = parser.add_argument_group("required")

    required.add_argument(
        "--snapshot_ids",
        nargs="+",
        help="EC2 snapshot IDs to move, separated by space.",
        required=True,
    )
    required.add_argument(
        "--target_region", help="Target region for the snapshots.", required=True
    )
    parser.add_argument(
        "--pending_limit",
        type=int,
        help="Limit for concurrent snapshot copy operations.",
        required=False,
    )
    parser.add_argument(
        "--encrypt",
        action="store_true",
        help="Encrypt or re-encrypt the snapshots before moving.",
        required=False,
    )
    parser.add_argument(
        "--kms_key_id",
        help="KMS Key ID to use for snapshot encryption.",
        required=False,
    )


def cli(args):
    """Command-line interface function"""

    inputs = vars(args)

    try:
        result = copy_snapshots(**inputs)
        set_output(result, inputs)
    except Exception as e:
        print(f"Action failed - {e}")
        exit(1)



from avtomat_aws.helpers.cli.set_output import set_output
from avtomat_aws.services.ec2 import modify_tags

ACTION_DESCRIPTION = "Modify EC2 resource tags."


def add_cli_arguments(parser):
    """Argument parsing"""

    required = parser.add_argument_group("required")

    required.add_argument(
        "--resource_ids",
        nargs="+",
        help="EC2 resource IDs to apply tags to, separated by space.",
        required=True,
    )
    required.add_argument(
        "--tags",
        nargs="+",
        help="Tags to create or delete in Key=Value format, separated by space.",
        required=True,
    )
    parser.add_argument(
        "--create", action="store_true", help="Create tags.", required=False
    )
    parser.add_argument(
        "--delete", action="store_true", help="Delete tags.", required=False
    )
    parser.add_argument(
        "--dynamic_tags",
        action="store_true",
        help="Add resource_id in the tag key or value.",
        required=False,
    )


def cli(args):
    """Command-line interface function"""

    inputs = vars(args)

    try:
        result = modify_tags(**inputs)
        set_output(result, inputs)
    except Exception as e:
        print(f"Action failed - {e}")
        exit(1)



from avtomat_aws.helpers.cli.set_output import set_output
from avtomat_aws.services.ec2 import delete_instances

ACTION_DESCRIPTION = (
    "Delete EC2 instances and, optionally, any associated resource types."
)


def add_cli_arguments(parser):
    """Argument parsing"""

    required = parser.add_argument_group("required")

    required.add_argument(
        "--instance_ids",
        nargs="+",
        help="Instance IDs to delete, separated by space.",
        required=True,
    )
    parser.add_argument(
        "--include",
        nargs="+",
        help="Associated resource types to delete, separated by space.",
        required=False,
    )
    parser.add_argument(
        "--final_image",
        action="store_true",
        help="Create an image (AMI) before deletion.",
        required=False,
    )
    parser.add_argument(
        "--disable_protections",
        action="store_true",
        help="Disable termination and stop protection before deletion.",
        required=False,
    )


def cli(args):
    """Command-line interface function"""

    inputs = vars(args)

    try:
        result = delete_instances(**inputs)
        set_output(result, inputs)
    except Exception as e:
        print(f"Action failed - {e}")
        exit(1)



import logging

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {"re_encrypt": False, "region": None, "debug": False, "silent": False}
RULES = [{"required": ["instance_id", "kms_key_id"]}]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def encrypt_instance_volumes(**kwargs):
    """Encrypt all instance volumes with a KMS key"""

    # Required parameters
    instance_id = kwargs.pop("instance_id")
    kms_key_id = kwargs.pop("kms_key_id")

    session_objects = set_session_objects(
        kwargs.get("session"),
        clients=["ec2"],
        resources=["ec2"],
        region=kwargs["region"],
    )

    logger.info(f"{instance_id} - encrypting volumes with {kms_key_id}")

    ec2_resource = session_objects["ec2_resource"]
    ec2_client = session_objects["ec2_client"]

    # Check instance state to ensure it's started only if originally running
    instance = ec2_resource.Instance(instance_id)
    originally_running = False
    if instance.state["Name"] == "running":
        originally_running = True

    # Create snapshots and collect volume details
    volume_details = []
    for volume in instance.volumes.all():
        if not kwargs.get("re_encrypt") and volume.encrypted:
            logger.info(f"{volume.id} - skipped, already encrypted")
            continue
        device = volume.attachments[0]["Device"]
        snapshot = session_objects["ec2_resource"].create_snapshot(
            VolumeId=volume.id,
            Description=f"Pre-encryption snapshot of {volume.id}",
            TagSpecifications=[
                {
                    "ResourceType": "snapshot",
                    "Tags": [{"Key": "Name", "Value": f"source:{volume.id}"}],
                }
            ],
        )
        logger.info(f"{snapshot.id} - creating from {volume.id}")
        volume_details.append((volume, device, snapshot))

    # Wait for snapshots to complete
    for _, _, snapshot in volume_details:
        snapshot.wait_until_completed()

    # Create encrypted volumes from snapshots
    encrypted_volumes_info = []
    for volume, device, snapshot in volume_details:
        create_volume_args = {
            "SnapshotId": snapshot.id,
            "AvailabilityZone": volume.availability_zone,
            "Encrypted": True,
            "KmsKeyId": kms_key_id,
        }
        if volume.tags:
            create_volume_args["TagSpecifications"] = [
                {"ResourceType": "volume", "Tags": volume.tags}
            ]

        encrypted_volume = ec2_resource.create_volume(**create_volume_args)
        encrypted_volumes_info.append((encrypted_volume, device))
        logger.info(f"{encrypted_volume.id} - creating")

    # Wait for all encrypted volumes to be available
    encrypted_volume_ids = [vol.id for vol, _ in encrypted_volumes_info]
    waiter = ec2_client.get_waiter("volume_available")

    # Proceed only if encrypted volumes have been created
    if encrypted_volume_ids:
        waiter.wait(VolumeIds=encrypted_volume_ids)

        # Stop instance
        instance.stop()
        logger.info(f"{instance_id} - stopping")
        instance.wait_until_stopped()

        # Detach original volumes
        for volume, _, _ in volume_details:
            volume.detach_from_instance()
            logger.info(f"{volume.id} ({volume.attachments[0]['Device']}) - detaching")

        # Wait for volumes to be detached
        waiter.wait(VolumeIds=[vol.id for vol, _, _ in volume_details])

        # Attach encrypted volumes
        for encrypted_volume, device in encrypted_volumes_info:
            instance.attach_volume(VolumeId=encrypted_volume.id, Device=device)
            logger.info(
                f"{encrypted_volume.id} - attaching to {instance.id} as {device}"
            )

        # Start instance
        if originally_running:
            instance.start()
            logger.info(f"{instance_id} - starting")
            instance.wait_until_running()

        # Delete snapshots
        for _, _, snapshot in volume_details:
            snapshot.delete()
            logger.info(f"{snapshot.id} - deleted")

        logger.info("To rollback:")
        logger.info("1. Attach")
        for volume, device, _ in volume_details:
            logger.info(f" -> {volume.id} to {instance_id} as {device}")

        logger.info("2. Delete")
        for encrypted_volume, _ in encrypted_volumes_info:
            logger.info(f" -> {encrypted_volume.id}")
    else:
        logger.info("No volumes to encrypt")

    return instance.id



from .table import output_table


def set_output(data, inputs):
    """Print the result of an action."""

    if inputs.get("output") == "table":
        output_table(data)
    else:
        print(data)



import logging

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.format_tags import format_tags
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {
    "create": False,
    "delete": False,
    "dynamic_tags": False,
    "region": None,
    "debug": False,
    "silent": False,
}
RULES = [
    {"required": ["resource_ids", "tags"]},
    {"at_least_one": ["create", "delete"]},
]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def modify_tags(**kwargs):
    """Create or delete tags for EC2 resources"""

    # Required parameters
    resource_ids = kwargs.pop("resource_ids")
    tags = kwargs.pop("tags")

    if not resource_ids:
        logger.info("No resources to modify tags for")
        return []

    session_objects = set_session_objects(
        kwargs["session"], clients=["ec2"], region=kwargs["region"]
    )

    if kwargs.get("create"):
        formatted_tags = format_tags(tags, creation=True)
        logger.info(f"Creating tags for resources")
    else:
        formatted_tags = format_tags(tags)
        logger.info(f"Deleting tags for resources")

    if kwargs.get("dynamic_tags"):
        failed_resources = dynamic_tags(
            formatted_tags, resource_ids, session_objects, **kwargs
        )
    else:
        failed_resources = static_tags(
            formatted_tags, resource_ids, session_objects, **kwargs
        )

    if failed_resources:
        logger.info(f"{len(failed_resources)} resources failed tag modification")
    else:
        logger.info("Success")
    logger.debug(failed_resources)

    return failed_resources


def dynamic_tags(tags, resource_ids, session_objects, **kwargs):
    """Create tags in dynamic mode, uses an API call per resource"""

    failed_resources = []
    for resource_id in resource_ids:
        # Replace {resource_id} with the actual resource_id
        dynamic_tags = []
        for tag in tags:
            dynamic_tags.append(
                {
                    "Key": tag["Key"].replace("{resource_id}", resource_id),
                    "Value": tag["Value"].replace("{resource_id}", resource_id),
                }
            )
        try:
            if kwargs.get("create"):
                session_objects["ec2_client"].create_tags(
                    Resources=[resource_id], Tags=dynamic_tags
                )
            elif kwargs.get("delete"):
                session_objects["ec2_client"].delete_tags(
                    Resources=[resource_id], Tags=dynamic_tags
                )
        except Exception as e:
            logger.error(f"Failed to modify tags for resource {resource_id} - {e}")
            failed_resources.append(resource_id)
            continue

    return failed_resources


def static_tags(tags, resource_ids, session_objects, **kwargs):
    """Create tags in static mode, uses an API call per 200 resources"""

    failed_resources = []
    if len(resource_ids) > 200:
        # Split the resources into chunks of 200
        chunks = [resource_ids[i : i + 200] for i in range(0, len(resource_ids), 200)]
        for chunk in chunks:
            try:
                if kwargs.get("create"):
                    session_objects["ec2_client"].create_tags(
                        Resources=chunk, Tags=tags
                    )
                elif kwargs.get("delete"):
                    session_objects["ec2_client"].delete_tags(
                        Resources=chunk, Tags=tags
                    )
            except Exception as e:
                logger.error(f"Failed to modify tags for resources for a chunk - {e}")
                failed_resources.extend(chunk)
                continue
    else:
        try:
            if kwargs.get("create"):
                session_objects["ec2_client"].create_tags(
                    Resources=resource_ids, Tags=tags
                )
            elif kwargs.get("delete"):
                session_objects["ec2_client"].delete_tags(
                    Resources=resource_ids, Tags=tags
                )
        except Exception as e:
            logger.error(f"Failed to modify tags for resources - {e}")
            failed_resources.extend(resource_ids)

    return failed_resources



from avtomat_aws.helpers.cli.set_output import set_output
from avtomat_aws.services.iam import discover_no_mfa_users

ACTION_DESCRIPTION = "Discover IAM users without MFA enabled."


def add_cli_arguments(parser):
    """Argument parsing"""
    pass


def cli(args):
    """Command-line interface function"""

    inputs = vars(args)

    try:
        result = discover_no_mfa_users(**inputs)
        set_output(result, inputs)
    except Exception as e:
        print(f"Action failed - {e}")
        exit(1)



import logging

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {"instance_ids": [], "region": None, "debug": False, "silent": False}


@validate(DEFAULTS)
@set_logger()
@authenticate()
def discover_no_ssm_instances(**kwargs):
    """Discover EC2 instances without SSM enabled"""

    session_objects = set_session_objects(
        kwargs["session"],
        clients=["ec2", "ssm"],
        resources=["ec2"],
        region=kwargs["region"],
    )

    logger.info(f"Discovering instances without SSM enabled")

    ssm_instances = []
    response = session_objects["ssm_client"].describe_instance_information(
        InstanceInformationFilterList=[{"key": "PingStatus", "valueSet": ["Online"]}]
    )
    while True:
        for instance in response["InstanceInformationList"]:
            ssm_instances.append(instance["InstanceId"])

        if response.get("NextToken"):
            response = session_objects["ssm_client"].describe_instance_information(
                InstanceInformationFilterList=[
                    {"key": "PingStatus", "valueSet": ["Online"]}
                ],
                NextToken=response["NextToken"],
            )
        else:
            break

    no_ssm_instances = []
    for instance in session_objects["ec2_resource"].instances.filter(
        InstanceIds=kwargs["instance_ids"]
    ):
        if instance.id not in ssm_instances:
            no_ssm_instances.append(instance.id)

    logger.info(f"{len(no_ssm_instances)} instances found")
    logger.debug(no_ssm_instances)

    return no_ssm_instances



import logging

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {"region": None, "debug": False, "silent": False}
RULES = [{"required": ["backup_vault_name", "recovery_point_arns"]}]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def delete_backups(**kwargs):
    """Delete backup recovery points"""

    # Required parameters
    backup_vault_name = kwargs.pop("backup_vault_name")
    recovery_point_arns = kwargs.pop("recovery_point_arns")

    if not recovery_point_arns:
        logger.info("No recovery points to delete")
        return []

    session_objects = set_session_objects(
        kwargs["session"], clients=["backup"], region=kwargs["region"]
    )

    logger.info(f"Deleting recovery points from '{backup_vault_name}' backup vault")

    failed = []
    counter = 0
    for arn in recovery_point_arns:
        try:
            session_objects["backup_client"].delete_recovery_point(
                BackupVaultName=backup_vault_name, RecoveryPointArn=arn
            )
            logger.info(f"{arn} - deleted")
            counter += 1
        except Exception as e:
            failed.append(arn)
            logger.error(f"{arn} - failed to delete - {e}")

    logger.info(f"{counter} recovery points deleted")
    if failed:
        logger.warning(f"{len(failed)} recovery points failed to delete")
        logger.debug(failed)

    return failed



import logging
import time
import uuid

from botocore.exceptions import ClientError, WaiterError

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {
    "include": None,
    "final_image": False,
    "disable_protections": False,
    "region": None,
    "debug": False,
    "silent": False,
}
RULES = [
    {"required": ["instance_ids"]},
    {
        "choice": [
            {"include": ["volume", "image", "snapshot"]},
        ]
    },
]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def delete_instances(**kwargs):
    """Delete EC2 instances"""

    # Required parameters
    instance_ids = kwargs.pop("instance_ids")

    if not instance_ids:
        logger.info("No instances to delete")
        return []

    session_objects = set_session_objects(
        kwargs["session"], resources=["ec2"], region=kwargs["region"]
    )
    failed = {"Instances": [], "Volumes": [], "Images": [], "Snapshots": []}

    # Get instance objects
    instances = []
    for instance_id in instance_ids:
        try:
            instance = session_objects["ec2_resource"].Instance(instance_id)
            useless = instance.state  # Check if instance exists
            instances.append(instance)
        except Exception as e:
            failed["Instances"].append(instance_id)
            logger.error(f"{instance_id} - failed to discover instance")
            continue

    # Gatekeeper: If no instances remain, return early
    if not instances:
        return failed

    # Discover associated resources before deletion
    associated_resources = {"Volumes": [], "Images": [], "Snapshots": []}
    if kwargs.get("include"):
        logger.info(f"Discovering associated resources: {kwargs.get('include')}")
        if "volume" in kwargs.get("include"):
            associated_resources["Volumes"] = discover_volumes(instances)
        if "image" in kwargs.get("include"):
            associated_resources["Images"], associated_resources["Snapshots"] = (
                discover_images(instances, session_objects["ec2_resource"])
            )
        if "snapshot" in kwargs.get("include"):
            if not associated_resources[
                "Volumes"
            ]:  # If volume deletion wasn't requested (we still need volumes for snapshots)
                associated_resources["Volumes"] = discover_volumes(instances)
            associated_resources["Snapshots"] += discover_snapshots(
                associated_resources["Volumes"], session_objects["ec2_resource"]
            )  # We append instead of assign to avoid overwriting snapshots associated with images
            associated_resources["Snapshots"] = list(
                set(associated_resources["Snapshots"])
            )  # Deduplicate

    if kwargs.get("disable_protections"):
        logger.info("Disabling API protections")
        disable_protections(instances)

    running_instances_to_stop = []
    logger.info("Stopping any running instances")
    for instance in instances:  # First concurrently stop all running instances.
        if instance.state["Name"] == "running":
            logger.info(f"{instance.id} - running, stopping before deletion")
            instance.stop()
            running_instances_to_stop.append(instance)
    for (
        instance
    ) in (
        running_instances_to_stop
    ):  # Then wait for them to stop. Avoids waiting per instance.
        instance.wait_until_stopped()

    images_to_wait = []
    if kwargs.get("final_image"):
        logger.info("Creating images before deletion")
        for instance in instances[:]:  # Iterate over a shallow copy
            try:
                # Generate random uuid as image names must be unique
                uuid_str = str(uuid.uuid4())
                image = instance.create_image(
                    Name=f"source_{instance.id}_{uuid_str}",
                    Description=f"Pre-deletion image of {instance.id}",
                    NoReboot=True,
                    TagSpecifications=[
                        {
                            "ResourceType": "image",
                            "Tags": instance.tags,
                        }
                    ],
                )
                logger.info(f"{instance.id} - creating {image.id}")
                images_to_wait.append((image, instance, uuid_str))
            except Exception as e:
                failed["Instances"].append(instance.id)
                instances.remove(instance)  # Modify the original list
                logger.error(
                    f"{instance.id} - failed to create image, will not delete - {e}"
                )
                continue

    # Gatekeeper: If no instances remain, return early
    if not instances:
        return failed

    for image, instance, _ in images_to_wait:
        try:
            image.wait_until_exists(
                Filters=[{"Name": "state", "Values": ["available"]}]
            )
        except WaiterError as e:
            failed["Instances"].append(instance.id)
            instances.remove(instance)
            logger.error(
                f"{instance.id} - failed to create image, will not delete - {e}"
            )
            continue

    # Gatekeeper: If no instances remain, return early
    if not instances:
        return failed

    logger.info(f"Deleting instances")
    for instance in instances[:]:  # Iterate over a shallow copy
        try:
            instance.terminate()
            logger.info(f"{instance.id} - deleted")
        except Exception as e:
            failed["Instances"].append(instance.id)
            instances.remove(instance)  # Modify the original list
            logger.error(f"{instance.id} - failed to delete - {e}")
            if kwargs.get(
                "final_image"
            ):  # Cleanup newly created image if deletion fails
                for image, _, uuid_str in images_to_wait:
                    if image.name == f"source_{instance.id}_{uuid_str}":
                        snapshots = discover_image_snapshots(
                            image, session_objects["ec2_resource"]
                        )
                        image.deregister()
                        delete_snapshots(snapshots, cleanup_final_image=True)
                        logger.warning(f"{instance.id} - cleaned up {image.id}")
            continue

    # Gatekeeper: If no instances remain, return early
    if not instances:
        return failed

    time.sleep(
        5
    )  # Wait for a few seconds before attempting to delete associated resources

    volumes_counter = 0
    images_counter = 0
    snapshots_counter = 0
    if kwargs.get("include"):
        logger.info(f"Deleting associated resources: {kwargs.get('include')}")
        if "volume" in kwargs.get("include"):
            volumes_counter, failed["Volumes"] = delete_volumes(
                associated_resources["Volumes"]
            )
        if "image" in kwargs.get("include"):
            images_counter, failed["Images"] = delete_images(
                associated_resources["Images"]
            )
        if "image" in kwargs.get("include") or "snapshot" in kwargs.get("include"):
            snapshots_counter, failed["Snapshots"] = delete_snapshots(
                associated_resources["Snapshots"]
            )

        # Provide a summary of deleted resources
        logger.info(f"{len(instances)} instances deleted")
        logger.info(f"{volumes_counter} volumes deleted")
        logger.info(f"{images_counter} images deleted")
        logger.info(f"{snapshots_counter} snapshots deleted")

        if failed["Instances"]:
            logger.warning(f"{len(failed['Instances'])} instances failed to delete")
            logger.debug(failed["Instances"])
        if failed["Volumes"]:
            logger.warning(f"{len(failed['Volumes'])} volumes failed to delete")
            logger.debug(failed["Volumes"])
        if failed["Images"]:
            logger.warning(f"{len(failed['Images'])} images failed to delete")
            logger.debug(failed["Images"])
        if failed["Snapshots"]:
            logger.warning(f"{len(failed['Snapshots'])} snapshots failed to delete")
            logger.debug(failed["Snapshots"])

    return failed


def delete_volumes(volumes):
    """Delete all volumes associated with an instance"""
    counter = 0
    failed = []
    for volume in volumes:
        try:
            volume.delete()
            counter += 1
            logger.info(f"{volume.id} - deleted")
        except Exception as e:
            failed.append(volume.id)
            logger.error(f"{volume.id} - failed to delete - {e}")
            continue
    return counter, failed


def delete_images(images):
    """Delete all images associated with an instance"""
    counter = 0
    failed = []
    for image in images:
        try:
            image.deregister()
            counter += 1
            logger.info(f"{image.id} - deleted")
        except Exception as e:
            failed.append(image.id)
            logger.error(f"{image.id} - failed to delete - {e}")
            continue
    return counter, failed


def delete_snapshots(snapshots, cleanup_final_image=False):
    """Delete all snapshots associated with an image"""
    counter = 0
    failed = []
    for snapshot in snapshots:
        try:
            snapshot.delete()
            counter += 1
            if (
                not cleanup_final_image
            ):  # Don't log extra messages if cleaning up a final image
                logger.info(f"{snapshot.id} - deleted")
        except Exception as e:
            failed.append(snapshot.id)
            logger.error(f"{snapshot.id} - failed to delete - {e}")
            continue
    return counter, failed


def discover_volumes(instances):
    """Discover volumes associated with instances"""
    volumes = []
    for instance in instances:
        volumes += instance.volumes.all()
    return volumes


def discover_images(instances, ec2_resource):
    """Discover images (and their underlying snapshots) associated with instances"""
    instance_ids = [instance.id for instance in instances]
    images = []
    for image in ec2_resource.images.filter(
        Owners=["self"],
        Filters=[{"Name": "source-instance-id", "Values": instance_ids}],
    ):
        images.append(image)
    snapshots = []
    for image in images:
        if image.state == "available":
            snapshots += discover_image_snapshots(image, ec2_resource)
        else:
            logger.warning(
                f"{image.id} - not fully available, will skip snapshot deletion"
            )
    return images, snapshots


def discover_snapshots(volumes, ec2_resource):
    """Discover snapshots associated with volumes"""
    volume_ids = [volume.id for volume in volumes]
    snapshots = []
    for snapshot in ec2_resource.snapshots.filter(
        OwnerIds=["self"], Filters=[{"Name": "volume-id", "Values": volume_ids}]
    ):
        snapshots.append(snapshot)
    return snapshots


def discover_image_snapshots(image, ec2_resource):
    """Discover snapshots associated with images"""
    snapshots = []
    for snapshot_id in image.block_device_mappings:
        try:
            snapshot = ec2_resource.Snapshot(snapshot_id["Ebs"]["SnapshotId"])
            snapshots.append(snapshot)
        except Exception as e:
            logger.error(f"{image.id} - failed to discover snapshots - {e}")
            continue
    return snapshots


def disable_protections(instances):
    """Disable termination and stop protections for instances"""
    for instance in instances:
        try:
            instance.modify_attribute(DisableApiTermination={"Value": False})
            instance.modify_attribute(DisableApiStop={"Value": False})
            logger.info(f"{instance.id} - api protections disabled")
        except Exception as e:
            logger.error(f"{instance.id} - failed to disable api protections - {e}")



from avtomat_aws.helpers.cli.set_output import set_output
from avtomat_aws.services.cloudtrail import discover_user_events

ACTION_DESCRIPTION = "Discover events created by specific user."


def add_cli_arguments(parser):
    """Argument parsing"""

    required = parser.add_argument_group("required")

    required.add_argument(
        "--user", help="Search events originating from this user.", required=True
    )
    parser.add_argument(
        "--events",
        nargs="+",
        help="Search for specific events, separated by space.",
        required=False,
    )
    parser.add_argument(
        "--created_after",
        help="Search for events created after this time. (YYYY/MM/DD)",
        required=False,
    )
    parser.add_argument(
        "--created_before",
        help="Search for events created before this time. (YYYY/MM/DD)",
        required=False,
    )


def cli(args):
    """Command-line interface function"""

    inputs = vars(args)

    try:
        result = discover_user_events(**inputs)
        set_output(result, inputs)
    except Exception as e:
        print(f"Action failed - {e}")
        exit(1)



import logging

from botocore.exceptions import ClientError

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {"region": None, "debug": False, "silent": False}
RULES = [{"required": ["user"]}]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def quarantine_user(**kwargs):
    """Disable console and programmatic access and apply AWSCompromisedKeyQuarantineV2 policy to an IAM user"""

    # Required parameters
    user = kwargs.pop("user")

    session_objects = set_session_objects(
        kwargs["session"], clients=["iam"], region=kwargs["region"]
    )

    logger.info(f"Quarantining user {user}")

    try:
        session_objects["iam_client"].delete_login_profile(UserName=user)
    except ClientError as e:
        if e.response["Error"]["Code"] == "NoSuchEntity":
            pass
        else:
            raise e
    logger.info(f"Console access disabled")

    keys_response = session_objects["iam_client"].list_access_keys(UserName=user)
    for key in keys_response["AccessKeyMetadata"]:
        if key["Status"] == "Active":
            session_objects["iam_client"].update_access_key(
                AccessKeyId=key["AccessKeyId"], Status="Inactive", UserName=user
            )
    logger.info(f"Programmatic access disabled")

    session_objects["iam_client"].attach_user_policy(
        UserName=user, PolicyArn="arn:aws:iam::aws:policy/AWSCompromisedKeyQuarantineV2"
    )
    logger.info(f"AWSCompromisedKeyQuarantineV2 policy attached")



import logging

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.format_tags import format_tags
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {
    "existing": False,
    "missing": False,
    "region": None,
    "debug": False,
    "silent": False,
}
RULES = [
    {"required": ["resource_types", "tags"]},
    {
        "choice": [
            {
                "resource_types": [
                    "image",
                    "instance",
                    "internet_gateway",
                    "key_pair",
                    "network_acl",
                    "route_table",
                    "security_group",
                    "snapshot",
                    "subnet",
                    "volume",
                    "vpc",
                ]
            },
        ]
    },
    {"at_most_one": ["existing", "missing"]},
    {"at_least_one": ["existing", "missing"]},
]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def discover_tags(**kwargs):
    """Discover specific tags that exist on EC2 resources"""

    # Required parameters
    resource_types = kwargs.pop("resource_types")
    tags = kwargs.pop("tags")

    session_objects = set_session_objects(
        kwargs["session"], resources=["ec2"], region=kwargs["region"]
    )

    if kwargs.get("existing"):
        logger.info(f"Discovering resources with tags: {tags}")
    elif kwargs.get("missing"):
        logger.info(f"Discovering resources without tags: {tags}")

    formatted_tags = format_tags(tags)
    resources = []
    if "image" in resource_types:
        response = session_objects["ec2_resource"].images.filter(Owners=["self"])
        resources.extend(search_collections(response, formatted_tags, **kwargs))
    if "instance" in resource_types:
        response = session_objects["ec2_resource"].instances.all()
        resources.extend(search_collections(response, formatted_tags, **kwargs))
    if "internet_gateway" in resource_types:
        response = session_objects["ec2_resource"].internet_gateways.all()
        resources.extend(search_collections(response, formatted_tags, **kwargs))
    if "key_pair" in resource_types:
        response = session_objects["ec2_resource"].key_pairs.all()
        resources.extend(search_collections(response, formatted_tags, **kwargs))
    if "network_acl" in resource_types:
        response = session_objects["ec2_resource"].network_acls.all()
        resources.extend(search_collections(response, formatted_tags, **kwargs))
    if "route_table" in resource_types:
        response = session_objects["ec2_resource"].route_tables.all()
        resources.extend(search_collections(response, formatted_tags, **kwargs))
    if "security_group" in resource_types:
        response = session_objects["ec2_resource"].security_groups.all()
        resources.extend(search_collections(response, formatted_tags, **kwargs))
    if "snapshot" in resource_types:
        response = session_objects["ec2_resource"].snapshots.filter(OwnerIds=["self"])
        resources.extend(search_collections(response, formatted_tags, **kwargs))
    if "subnet" in resource_types:
        response = session_objects["ec2_resource"].subnets.all()
        resources.extend(search_collections(response, formatted_tags, **kwargs))
    if "volume" in resource_types:
        response = session_objects["ec2_resource"].volumes.all()
        resources.extend(search_collections(response, formatted_tags, **kwargs))
    if "vpc" in resource_types:
        response = session_objects["ec2_resource"].vpcs.all()
        resources.extend(search_collections(response, formatted_tags, **kwargs))

    logger.info(f"{len(resources)} resources found")
    logger.debug(resources)

    return resources


def search_collections(response, tags, **kwargs):
    """Iterate through resources in the response and return ones that have or don't have supplied tags"""

    missing = kwargs.get("missing")
    existing = kwargs.get("existing")

    resources = []
    for resource in response:
        resource_tags = {
            tag["Key"]: tag.get("Value", None)
            for tag in getattr(resource, "tags") or []
        }
        match = False

        if missing:
            match = any(
                (
                    tag["Key"] not in resource_tags
                    or (
                        tag.get("Value") is not None
                        and resource_tags[tag["Key"]] != tag["Value"]
                    )
                )
                for tag in tags
            )

        elif existing:
            match = all(
                tag["Key"] in resource_tags
                and (
                    tag.get("Value") is None
                    or resource_tags[tag["Key"]] == tag["Value"]
                )
                for tag in tags
            )

        if match:
            try:
                resources.append(resource.id)
            except AttributeError:
                resources.append(resource.name)

    return resources



import logging

import boto3

logger = logging.getLogger(__name__)


def set_session(**kwargs):
    """Set the session."""

    # Set logging level
    if kwargs.get("debug"):
        logger.setLevel(logging.DEBUG)
    if kwargs.get("silent"):
        logger.setLevel(logging.WARNING)

    if kwargs.get("session"):
        logger.debug("Session explicitly supplied.")
        session = kwargs["session"]
    elif kwargs.get("profile"):
        session = authenticate_profile(kwargs["profile"])
    elif kwargs.get("role_arn"):
        session = authenticate_role(
            kwargs["role_arn"], kwargs["mfa_serial"], kwargs["mfa_token"]
        )
    elif kwargs.get("access_key") and kwargs.get("secret_key"):
        session = authenticate_credentials(
            kwargs["access_key"], kwargs["secret_key"], kwargs["session_token"]
        )
    else:
        logger.debug(
            "No 'session' or 'profile' or 'role_arn' or 'access_key' + 'secret_key' parameters supplied. "
            "Falling back to default credentials."
        )
        session = authenticate_default()

    sts_client = session.client("sts")
    response = sts_client.get_caller_identity()
    logger.debug(f"Session created: {response['Arn']}")

    return session


def authenticate_credentials(access_key, secret_key, session_token=None):
    """Authenticate with credentials"""

    return boto3.Session(
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key,
        aws_session_token=session_token,
    )


def authenticate_profile(profile_name):
    """Authenticate with a profile"""

    return boto3.Session(profile_name=profile_name)


def authenticate_role(role_arn, mfa_serial=None, mfa_token=None):
    """Authenticate by assuming a role"""

    sts_client = boto3.client("sts")
    response = sts_client.get_caller_identity()
    if ":" in response["UserId"]:
        parts = response["UserId"].split(":")
        user = parts[1]
    else:
        user = response["UserId"]

    assume_role_kwargs = {
        "RoleArn": role_arn,
        "RoleSessionName": user,
    }

    if mfa_serial:
        assume_role_kwargs["SerialNumber"] = mfa_serial
        assume_role_kwargs["TokenCode"] = mfa_token

    response = sts_client.assume_role(**assume_role_kwargs)
    credentials = response["Credentials"]

    return boto3.Session(
        aws_access_key_id=credentials["AccessKeyId"],
        aws_secret_access_key=credentials["SecretAccessKey"],
        aws_session_token=credentials["SessionToken"],
    )


def authenticate_default():
    """Authenticate with default credentials"""

    return boto3.Session()



from avtomat_aws.helpers.cli.set_output import set_output
from avtomat_aws.services.ec2 import encrypt_instance_volumes

ACTION_DESCRIPTION = "Encrypt all EBS volumes attached to an instance."


def add_cli_arguments(parser):
    """Argument parsing"""

    required = parser.add_argument_group("required")

    required.add_argument(
        "--instance_id", help="Instance ID for volume encryption.", required=True
    )
    required.add_argument(
        "--kms_key_id", help="KMS Key ID to use for encryption.", required=True
    )
    parser.add_argument(
        "--re_encrypt",
        action="store_true",
        help="Re-encrypt already encrypted volumes.",
        required=False,
    )


def cli(args):
    """Command-line interface function"""

    inputs = vars(args)

    try:
        result = encrypt_instance_volumes(**inputs)
        set_output(result, inputs)
    except Exception as e:
        print(f"Action failed - {e}")
        exit(1)



import logging
from datetime import datetime, timezone

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {
    "image_ids": [],
    "public": False,
    "exclude_aws_backup": False,
    "created_before": None,
    "created_after": None,
    "region": None,
    "debug": False,
    "silent": False,
}
RULES = [{"date_yymmdd": ["created_before", "created_after"]}]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def discover_images(**kwargs):
    """Discover images based on provided criteria"""

    if kwargs.get("created_after") and type(kwargs["created_after"]) is str:
        kwargs["created_after"] = datetime.strptime(kwargs["created_after"], "%Y/%m/%d")
    if kwargs.get("created_before") and type(kwargs["created_before"]) is str:
        kwargs["created_before"] = datetime.strptime(
            kwargs["created_before"], "%Y/%m/%d"
        )

    logger.info("Discovering images")

    filters = build_filters(**kwargs)
    images = search_images(filters, **kwargs)

    logger.info(f"{len(images)} images found")
    logger.debug(images)

    return images


def build_filters(**kwargs):
    """Construct filters"""

    image_ids = kwargs.get("image_ids")
    public = kwargs.get("public")
    exclude_aws_backup = kwargs.get("exclude_aws_backup")

    filters = []
    if public:
        logger.debug("Filtering for public images")
        filters.append({"Name": "is-public", "Values": ["true"]})
    if exclude_aws_backup:
        logger.debug("Filtering out AWS Backup images")
    if image_ids:
        logger.debug(f"Filtering for images: {image_ids}")

    return filters


def search_images(filters, **kwargs):
    """Search for images in specified region"""

    session = kwargs["session"]
    region = kwargs["region"]
    image_ids = kwargs.get("image_ids")
    created_before = kwargs.get("created_before")
    created_after = kwargs.get("created_after")
    exclude_aws_backup = kwargs.get("exclude_aws_backup")

    session_objects = set_session_objects(session, resources=["ec2"], region=region)
    images = []

    response = session_objects["ec2_resource"].images.filter(
        Filters=filters, Owners=["self"], ImageIds=image_ids
    )

    for image in response:
        creation_date = datetime.fromisoformat(
            image.creation_date.replace("Z", "+00:00")
        )
        if created_before and creation_date > created_before.replace(
            tzinfo=timezone.utc
        ):
            continue
        if created_after and creation_date < created_after.replace(tzinfo=timezone.utc):
            continue
        if exclude_aws_backup and any(
            tag["Key"].startswith("aws:backup") for tag in (image.tags or [])
        ):
            continue
        images.append(image.id)

    return images



from avtomat_aws.helpers.cli.set_output import set_output
from avtomat_aws.services.ec2 import discover_images

ACTION_DESCRIPTION = "Discover AWS images (AMI)."


def add_cli_arguments(parser):
    """Argument parsing"""

    parser.add_argument(
        "--image_ids",
        nargs="+",
        help="Image IDs to focus on, separated by space.",
        required=False,
    )
    parser.add_argument(
        "--public",
        action="store_true",
        help="Get only publicly exposed images.",
        required=False,
    )
    parser.add_argument(
        "--exclude_aws_backup",
        action="store_true",
        help="Exclude images managed by AWS Backup.",
        required=False,
    )
    parser.add_argument(
        "--created_before",
        help="Get images created before date (YYYY/MM//DD).",
        required=False,
    )
    parser.add_argument(
        "--created_after",
        help="Get images created after date (YYYY/MM//DD).",
        required=False,
    )


def cli(args):
    """Command-line interface function"""

    inputs = vars(args)

    try:
        result = discover_images(**inputs)
        set_output(result, inputs)
    except Exception as e:
        print(f"Action failed - {e}")
        exit(1)



import logging

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {"include_snapshots": False, "region": None, "debug": False, "silent": False}
RULES = [{"required": ["image_ids"]}]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def delete_images(**kwargs):
    """Delete EC2 images (AMI)"""

    # Required parameters
    image_ids = kwargs.pop("image_ids")

    if not image_ids:
        logger.info("No images to delete")
        return []

    session_objects = set_session_objects(
        kwargs["session"], resources=["ec2"], region=kwargs["region"]
    )

    if kwargs.get("include_snapshots"):
        logger.info(f"Deleting images and associated snapshots")
    else:
        logger.info(f"Deleting images")

    snapshot_ids = []
    failed = []
    counter = 0
    for image_id in image_ids:
        try:
            image = session_objects["ec2_resource"].Image(image_id)
            if kwargs.get("include_snapshots"):
                for block_device in image.block_device_mappings:
                    if "Ebs" in block_device:
                        snapshot_ids.append(block_device["Ebs"]["SnapshotId"])
            image.deregister()
            counter += 1
            logger.info(f"{image_id} - deleted")
        except Exception as e:
            failed.append(image_id)
            logger.error(f"{image_id} - failed to delete - {e}")

    if snapshot_ids:
        delete_snapshots(snapshot_ids, session_objects)

    logger.info(f"{counter} images deleted")
    if failed:
        logger.warning(f"{len(failed)} images failed to delete")
        logger.debug(failed)

    return failed


def delete_snapshots(snapshot_ids, session_objects):
    """Delete EBS snapshots"""

    for snapshot_id in snapshot_ids:
        try:
            snapshot = session_objects["ec2_resource"].Snapshot(snapshot_id)
            snapshot.delete()
            logger.debug(f"{snapshot_id} - deleted")
        except Exception as e:
            logger.error(f"{snapshot_id} - failed to delete - {e}")



import logging

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {"region": None, "debug": False, "silent": False}


@validate(DEFAULTS)
@set_logger()
@authenticate()
def discover_unused_security_groups(**kwargs):
    """Discover unused security groups"""

    session = kwargs["session"]
    region = kwargs["region"]

    logger.info("Discovering unused security groups")

    all_sgs = list_all_sgs(session, region)

    attached_sgs = set()
    attached_sgs.update(list_eni_attached_sgs(session, region))
    attached_sgs.update(list_ec2_attached_sgs(session, region))
    attached_sgs.update(list_rds_attached_sgs(session, region))
    attached_sgs.update(list_elb_attached_sgs(session, region))
    attached_sgs.update(list_elbv2_attached_sgs(session, region))
    attached_sgs.update(list_lambda_attached_sgs(session, region))
    attached_sgs.update(list_redshift_attached_sgs(session, region))
    attached_sgs.update(list_elasticache_attached_sgs(session, region))
    attached_sgs.update(list_emr_attached_sgs(session, region))
    attached_sgs.update(list_ecs_attached_sgs(session, region))
    attached_sgs.update(list_neptune_attached_sgs(session, region))
    attached_sgs.update(
        list_opensearch_attached_sgs(session, region)
    )  # No pagination support
    attached_sgs.update(list_msk_attached_sgs(session, region))

    unattached_sgs = list(all_sgs - attached_sgs)

    logger.info(f"{len(unattached_sgs)} unattached security groups found")
    logger.debug(unattached_sgs)

    return unattached_sgs


def list_all_sgs(session, region):
    """Return all security groups"""

    session_objects = set_session_objects(session, clients=["ec2"], region=region)
    sgs = set()

    response = session_objects["ec2_client"].describe_security_groups(MaxResults=500)

    if response:
        for sg in response["SecurityGroups"]:
            sgs.add(sg["GroupId"])

    while "NextToken" in response:
        response = session_objects["ec2_client"].describe_security_groups(
            MaxResults=500, NextToken=response["NextToken"]
        )
        for sg in response["SecurityGroups"]:
            sgs.add(sg["GroupId"])

    return sgs


def list_eni_attached_sgs(session, region):
    """Return security groups attached to ENIs"""

    session_objects = set_session_objects(session, clients=["ec2"], region=region)
    sgs = set()

    logger.info("Service: ENI")
    response = session_objects["ec2_client"].describe_network_interfaces(MaxResults=250)

    if response:
        for eni in response["NetworkInterfaces"]:
            for sg in eni.get("Groups", []):
                sgs.add(sg["GroupId"])

    while "NextToken" in response:
        response = session_objects["ec2_client"].describe_network_interfaces(
            MaxResults=250, NextToken=response["NextToken"]
        )
        for eni in response["NetworkInterfaces"]:
            for sg in eni.get("Groups", []):
                sgs.add(sg["GroupId"])

    return sgs


def list_ec2_attached_sgs(session, region):
    """Return security groups attached to EC2 instances"""

    session_objects = set_session_objects(session, clients=["ec2"], region=region)
    sgs = set()

    logger.info("Service: EC2")
    response = session_objects["ec2_client"].describe_instances(MaxResults=250)

    if response:
        for reservation in response["Reservations"]:
            for instance in reservation["Instances"]:
                for sg in instance["SecurityGroups"]:
                    sgs.add(sg["GroupId"])

    while "NextToken" in response:
        response = session_objects["ec2_client"].describe_instances(
            MaxResults=250, NextToken=response["NextToken"]
        )
        for reservation in response["Reservations"]:
            for instance in reservation["Instances"]:
                for sg in instance["SecurityGroups"]:
                    sgs.add(sg["GroupId"])

    return sgs


def list_rds_attached_sgs(session, region):
    """Return security groups attached to RDS instances"""

    session_objects = set_session_objects(session, clients=["rds"], region=region)
    sgs = set()

    logger.info("Service: RDS")
    response = session_objects["rds_client"].describe_db_instances(MaxRecords=100)

    if response:
        for instance in response["DBInstances"]:
            for sg in instance["VpcSecurityGroups"]:
                sgs.add(sg["VpcSecurityGroupId"])

    while "Marker" in response:
        response = session_objects["rds_client"].describe_db_instances(
            MaxRecords=100, Marker=response["Marker"]
        )
        for instance in response["DBInstances"]:
            for sg in instance["VpcSecurityGroups"]:
                sgs.add(sg["VpcSecurityGroupId"])

    return sgs


def list_elb_attached_sgs(session, region):
    """Return security groups attached to ELBs"""

    session_objects = set_session_objects(session, clients=["elb"], region=region)
    sgs = set()

    logger.info("Service: ELB")
    response = session_objects["elb_client"].describe_load_balancers(PageSize=400)

    if response:
        for lb in response["LoadBalancerDescriptions"]:
            for sg in lb.get("SecurityGroups", []):
                sgs.add(sg)

    while "NextMarker" in response:
        response = session_objects["elb_client"].describe_load_balancers(
            PageSize=400, Marker=response["NextMarker"]
        )
        for lb in response["LoadBalancerDescriptions"]:
            for sg in lb.get("SecurityGroups", []):
                sgs.add(sg)

    return sgs


def list_elbv2_attached_sgs(session, region):
    """Return security groups attached to ELBv2s"""

    session_objects = set_session_objects(session, clients=["elbv2"], region=region)
    sgs = set()

    logger.info("Service: ELBv2")
    response = session_objects["elbv2_client"].describe_load_balancers(PageSize=400)

    if response:
        for lb in response["LoadBalancers"]:
            for sg in lb.get("SecurityGroups", []):
                sgs.add(sg)

    while "NextMarker" in response:
        response = session_objects["elbv2_client"].describe_load_balancers(
            PageSize=400, Marker=response["NextMarker"]
        )
        for lb in response["LoadBalancers"]:
            for sg in lb.get("SecurityGroups", []):
                sgs.add(sg)

    return sgs


def list_lambda_attached_sgs(session, region):
    """Return security groups attached to Lambdas"""

    session_objects = set_session_objects(session, clients=["lambda"], region=region)
    sgs = set()

    logger.info("Service: Lambda")
    response = session_objects["lambda_client"].list_functions(MaxItems=50)

    if response:
        for function in response["Functions"]:
            for sg in function.get("VpcConfig", {}).get("SecurityGroupIds", []):
                sgs.add(sg)

    while "NextMarker" in response:
        response = session_objects["lambda_client"].list_functions(
            MaxItems=50, Marker=response["NextMarker"]
        )
        for function in response["Functions"]:
            for sg in function.get("VpcConfig", {}).get("SecurityGroupIds", []):
                sgs.add(sg)

    return sgs


def list_redshift_attached_sgs(session, region):
    """Return security groups attached to Redshift clusters"""

    session_objects = set_session_objects(session, clients=["redshift"], region=region)
    sgs = set()

    logger.info("Service: Redshift")
    response = session_objects["redshift_client"].describe_clusters(MaxRecords=100)

    if response:
        for cluster in response["Clusters"]:
            sg_response = cluster.get("VpcSecurityGroups")
            if sg_response:
                for sg in sg_response:
                    sgs.add(sg["VpcSecurityGroupId"])

    while "Marker" in response:
        response = session_objects["redshift_client"].describe_clusters(
            MaxRecords=100, Marker=response["Marker"]
        )
        for cluster in response["Clusters"]:
            sg_response = cluster.get("VpcSecurityGroups")
            if sg_response:
                for sg in sg_response:
                    sgs.add(sg["VpcSecurityGroupId"])

    return sgs


def list_elasticache_attached_sgs(session, region):
    """Return security groups attached to ElastiCache clusters"""

    session_objects = set_session_objects(
        session, clients=["elasticache"], region=region
    )
    sgs = set()

    logger.info("Service: ElastiCache")
    response = session_objects["elasticache_client"].describe_cache_clusters(
        MaxRecords=100
    )

    if response:
        for cluster in response["CacheClusters"]:
            sg_ids = cluster.get("SecurityGroups")
            if sg_ids:
                sgs.update(sg_ids)

    while "Marker" in response:
        response = session_objects["elasticache_client"].describe_cache_clusters(
            MaxRecords=100, Marker=response["Marker"]
        )
        for cluster in response["CacheClusters"]:
            sg_ids = cluster.get("SecurityGroups")
            if sg_ids:
                sgs.update(sg_ids)

    return sgs


def list_emr_attached_sgs(session, region):
    """Return security groups attached to EMR clusters"""

    session_objects = set_session_objects(session, clients=["emr"], region=region)
    sgs = set()

    logger.info("Service: EMR")
    response = session_objects["emr_client"].list_clusters(
        ClusterStates=["STARTING", "BOOTSTRAPPING", "RUNNING", "WAITING"]
    )

    if response:
        for cluster in response["Clusters"]:
            cluster_id = cluster["Id"]
            cluster_response = session_objects["emr_client"].describe_cluster(
                ClusterId=cluster_id
            )

            sg_ids = cluster_response["Cluster"]["Ec2InstanceAttributes"].get(
                "EmrManagedMasterSecurityGroup"
            )
            if sg_ids:
                sgs.add(sg_ids)
            sg_ids = cluster_response["Cluster"]["Ec2InstanceAttributes"].get(
                "EmrManagedSlaveSecurityGroup"
            )
            if sg_ids:
                sgs.add(sg_ids)
            sg_ids = cluster_response["Cluster"]["Ec2InstanceAttributes"].get(
                "ServiceAccessSecurityGroup"
            )
            if sg_ids:
                sgs.add(sg_ids)
            sg_ids = cluster_response["Cluster"]["Ec2InstanceAttributes"].get(
                "AdditionalMasterSecurityGroups"
            )
            if sg_ids:
                sgs.update(sg_ids)
            sg_ids = cluster_response["Cluster"]["Ec2InstanceAttributes"].get(
                "AdditionalSlaveSecurityGroups"
            )
            if sg_ids:
                sgs.update(sg_ids)

    while "Marker" in response:
        response = session_objects["emr_client"].list_clusters(
            ClusterStates=["STARTING", "BOOTSTRAPPING", "RUNNING", "WAITING"],
            Marker=response["Marker"],
        )
        for cluster in response["Clusters"]:
            cluster_id = cluster["Id"]
            cluster_response = session_objects["emr_client"].describe_cluster(
                ClusterId=cluster_id
            )

            sg_ids = cluster_response["Cluster"]["Ec2InstanceAttributes"].get(
                "EmrManagedMasterSecurityGroup"
            )
            if sg_ids:
                sgs.add(sg_ids)
            sg_ids = cluster_response["Cluster"]["Ec2InstanceAttributes"].get(
                "EmrManagedSlaveSecurityGroup"
            )
            if sg_ids:
                sgs.add(sg_ids)
            sg_ids = cluster_response["Cluster"]["Ec2InstanceAttributes"].get(
                "ServiceAccessSecurityGroup"
            )
            if sg_ids:
                sgs.add(sg_ids)
            sg_ids = cluster_response["Cluster"]["Ec2InstanceAttributes"].get(
                "AdditionalMasterSecurityGroups"
            )
            if sg_ids:
                sgs.update(sg_ids)
            sg_ids = cluster_response["Cluster"]["Ec2InstanceAttributes"].get(
                "AdditionalSlaveSecurityGroups"
            )
            if sg_ids:
                sgs.update(sg_ids)

    return sgs


def list_ecs_attached_sgs(session, region):
    """Return security groups attached to ECS tasks"""

    session_objects = set_session_objects(session, clients=["ecs"], region=region)
    sgs = set()

    logger.info("Service: ECS")

    # Paginate through ECS clusters
    def list_clusters():
        next_token = None
        while True:
            request_args = {"maxResults": 100}
            if next_token:
                request_args["nextToken"] = next_token
            response = session_objects["ecs_client"].list_clusters(**request_args)
            for cluster_arn in response["clusterArns"]:
                yield cluster_arn
            next_token = response.get("nextToken")
            if not next_token:
                break

    # Paginate through ECS tasks to collect SGs
    def list_tasks_and_collect_sgs(cluster_arn):
        next_token = None
        while True:
            request_args = {"cluster": cluster_arn, "maxResults": 100}
            if next_token:
                request_args["nextToken"] = next_token

            tasks_response = session_objects["ecs_client"].list_tasks(**request_args)

            task_arns = tasks_response.get("taskArns", [])
            if task_arns:
                described_tasks_response = session_objects["ecs_client"].describe_tasks(
                    cluster=cluster_arn, tasks=task_arns
                )
                described_tasks = described_tasks_response.get("tasks", [])
                for task in described_tasks:
                    for attachment in task.get("attachments", []):
                        for detail in attachment.get("details", []):
                            if detail.get("name") == "securityGroup":
                                sgs.add(detail["value"])

            next_token = tasks_response.get("nextToken")
            if not next_token:
                break

    for arn in list_clusters():
        list_tasks_and_collect_sgs(arn)

    return sgs


def list_neptune_attached_sgs(session, region):
    """Return security groups attached to Neptune instances"""

    session_objects = set_session_objects(session, clients=["neptune"], region=region)
    sgs = set()

    logger.info("Service: Neptune")
    response = session_objects["neptune_client"].describe_db_instances(MaxRecords=100)

    if response:
        for instance in response["DBInstances"]:
            for sg in instance["VpcSecurityGroups"]:
                sgs.add(sg["VpcSecurityGroupId"])

    while "Marker" in response:
        response = session_objects["neptune_client"].describe_db_instances(
            MaxRecords=100, Marker=response["Marker"]
        )
        for instance in response["DBClusters"]:
            for sg in instance["VpcSecurityGroups"]:
                sgs.add(sg["VpcSecurityGroupId"])

    return sgs


def list_opensearch_attached_sgs(session, region):
    """Return security groups attached to OpenSearch clusters"""

    session_objects = set_session_objects(
        session, clients=["opensearch"], region=region
    )
    sgs = set()

    logger.info("Service: OpenSearch")
    response = session_objects["opensearch_client"].list_domain_names()

    if response:
        for domain in response["DomainNames"]:
            domain_name = domain["DomainName"]
            domain_response = session_objects["opensearch_client"].describe_domain(
                DomainName=domain_name
            )
            sg_ids = domain_response["DomainStatus"]["VPCOptions"]["SecurityGroupIds"]
            sgs.update(sg_ids)

    return sgs


def list_msk_attached_sgs(session, region):
    """Return security groups attached to MSK clusters"""

    session_objects = set_session_objects(session, clients=["kafka"], region=region)
    sgs = set()

    logger.info("Service: MSK")
    response = session_objects["kafka_client"].list_clusters(MaxResults=100)

    if response:
        for cluster in response["ClusterInfoList"]:
            sgs.update(cluster["BrokerNodeGroupInfo"]["SecurityGroups"])

    while "NextToken" in response:
        response = session_objects["kafka_client"].list_clusters(
            MaxResults=100, NextToken=response["NextToken"]
        )
        for cluster in response["ClusterInfoList"]:
            sgs.update(cluster["BrokerNodeGroupInfo"]["SecurityGroups"])

    return sgs



from avtomat_aws.helpers.set_defaults import set_defaults
from avtomat_aws.validations.validate import validate as perform_validation


def validate(defaults=None, param_rules=None):
    """Decorator to apply default values and validate input parameters for the action."""

    defaults = defaults or {}
    param_rules = param_rules or []

    def decorator(func):
        def wrapper(*args, **kwargs):
            # Set defaults,supplied kwargs take precedence
            kwargs = set_defaults(kwargs, defaults)
            # Validate input parameters
            perform_validation(kwargs, param_rules)
            # Call the action
            return func(*args, **kwargs)

        return wrapper

    return decorator



def format_tags(tags, creation=False):
    """Format tags into a list of dictionaries"""
    formatted_tags = []
    # Format tags for creation
    # AWS doesn't accept None and requires at least an empty string for Value
    if creation:
        for tag in tags:
            if "=" in tag:
                key, value = tag.split("=", 1)
                formatted_tags.append({"Key": key, "Value": value})
            else:
                key = tag
                formatted_tags.append({"Key": key, "Value": ""})
    # Format tags for discovery or deletion
    else:
        for tag in tags:
            if "=" in tag:
                key, value = tag.split("=", 1)
                formatted_tags.append({"Key": key, "Value": value})
            else:
                key = tag
                formatted_tags.append({"Key": key})
    return formatted_tags



import logging

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {"prefix": "", "region": None, "debug": False, "silent": False}
RULES = [{"required": ["bucket", "objects"]}]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def delete_objects(**kwargs):
    """Delete objects from an S3 bucket"""

    # Required parameters
    bucket = kwargs.pop("bucket")
    objects = kwargs.pop("objects")

    if not objects:
        logger.info("No objects to delete")
        return []

    session_objects = set_session_objects(
        kwargs["session"], clients=["s3"], region=kwargs["region"]
    )

    logger.info(f"Deleting objects from '{bucket}'")

    if len(objects) > 500:
        chunks = [objects[x : x + 500] for x in range(0, len(objects), 500)]
    else:
        chunks = [objects]

    failed = []
    counter = 0
    for chunk in chunks:
        if kwargs.get("prefix"):
            prefix = kwargs["prefix"].rstrip("/")
            delete = {"Objects": [{"Key": f"{prefix}/{obj}"} for obj in chunk]}
        else:
            delete = {"Objects": [{"Key": obj} for obj in chunk]}
        try:
            response = session_objects["s3_client"].delete_objects(
                Bucket=bucket, Delete=delete
            )
            counter += len(response["Deleted"])
            if response.get("Errors"):
                for obj in response["Errors"]:
                    logger.error(
                        f"Failed to delete object: {obj['Key']} - {obj['Code']} - {obj['Message']}"
                    )
                    failed.extend(obj["Key"])
        except Exception as e:
            logger.error(f"Failed to delete objects: {e}")
            failed.extend(chunk)
            continue

    logger.info(f"{counter} objects deleted")
    if failed:
        logger.warning(f"{len(failed)} objects failed to delete")
        logger.debug(failed)

    return failed



import logging
from datetime import datetime, timezone

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {
    "volume_ids": [],
    "snapshot_ids": [],
    "unencrypted": False,
    "exclude_aws_backup": False,
    "created_before": None,
    "created_after": None,
    "region": None,
    "debug": False,
    "silent": False,
}
RULES = [
    {"date_yymmdd": ["created_before", "created_after"]},
    {"at_most_one": ["snapshot_ids", "volume_ids"]},
]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def discover_snapshots(**kwargs):
    """Discover snapshots based on provided criteria"""

    if kwargs.get("created_after") and type(kwargs["created_after"]) is str:
        kwargs["created_after"] = datetime.strptime(kwargs["created_after"], "%Y/%m/%d")
    if kwargs.get("created_before") and type(kwargs["created_before"]) is str:
        kwargs["created_before"] = datetime.strptime(
            kwargs["created_before"], "%Y/%m/%d"
        )

    logger.info("Discovering snapshots")

    filters = build_filters(**kwargs)
    snapshots = search_snapshots(filters, **kwargs)

    logger.info(f"{len(snapshots)} snapshots found")
    logger.debug(snapshots)

    return snapshots


def build_filters(**kwargs):
    """Construct filters"""

    unencrypted = kwargs.get("unencrypted")
    volume_ids = kwargs.get("volume_ids")
    snapshot_ids = kwargs.get("snapshot_ids")
    exclude_aws_backup = kwargs.get("exclude_aws_backup")

    filters = []
    if unencrypted:
        logger.debug("Filtering for unencrypted snapshots")
        filters.append({"Name": "encrypted", "Values": ["false"]})
    if volume_ids:
        logger.debug(f"Filtering for source volumes: {volume_ids}")
        filters.append({"Name": "volume-id", "Values": volume_ids})
    if exclude_aws_backup:
        logger.debug("Filtering out AWS Backup snapshots")
    if snapshot_ids:
        logger.debug(f"Filtering for snapshots: {snapshot_ids}")

    return filters


def search_snapshots(filters, **kwargs):
    """Search for snapshots in specified region"""

    session = kwargs["session"]
    region = kwargs["region"]
    snapshot_ids = kwargs.get("snapshot_ids")
    created_before = kwargs.get("created_before")
    created_after = kwargs.get("created_after")
    exclude_aws_backup = kwargs.get("exclude_aws_backup")

    session_objects = set_session_objects(session, resources=["ec2"], region=region)
    snapshots = []

    response = session_objects["ec2_resource"].snapshots.filter(
        Filters=filters, OwnerIds=["self"], SnapshotIds=snapshot_ids
    )

    for snapshot in response:
        if created_before and snapshot.start_time > created_before.replace(
            tzinfo=timezone.utc
        ):
            continue
        if created_after and snapshot.start_time < created_after.replace(
            tzinfo=timezone.utc
        ):
            continue
        if exclude_aws_backup and any(
            tag["Key"].startswith("aws:backup") for tag in (snapshot.tags or [])
        ):
            continue
        snapshots.append(snapshot.id)

    return snapshots



import logging
import os

logger = logging.getLogger(__name__)


def set_region(region=None, session=None, debug=None, silent=None):
    """Set the region."""

    # Set logging level
    if debug:
        logger.setLevel(logging.DEBUG)
    elif silent:
        logger.setLevel(logging.WARNING)

    if region:
        logger.debug(f"Region explicitly set")
    elif "AWS_DEFAULT_REGION" in os.environ:
        region = os.environ["AWS_DEFAULT_REGION"]
        logger.debug(f"Region found in environment variables")
    elif session.region_name:
        region = session.region_name
        logger.debug(f"Region found in configuration")
    else:
        region = "us-east-1"
        logger.debug(f"No session default, fallback to 'us-east-1'")

    logger.debug(f"Region - {region}")

    return region



import logging
from datetime import datetime, timezone

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {"region": None, "debug": False, "silent": False}
RULES = [{"required": ["threshold_days"]}]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def discover_unused_roles(**kwargs):
    """Discover IAM roles not used over a certain amount of days"""

    # Required parameters
    threshold_days = kwargs.pop("threshold_days")

    session_objects = set_session_objects(
        kwargs["session"], clients=["iam"], region=kwargs["region"]
    )

    logger.info(f"Discovering roles not used for more than {threshold_days} days")

    roles = []
    roles_response = session_objects["iam_client"].list_roles(MaxItems=100)
    while True:
        for role in roles_response["Roles"]:
            try:
                last_used = (
                    session_objects["iam_client"]
                    .get_role(RoleName=role["RoleName"])["Role"]
                    .get("RoleLastUsed")
                )
            except Exception as e:
                logger.error(
                    f"Failed to get last used date for {role['RoleName']} - {e}"
                )
                continue
            if last_used:
                age = (datetime.now(timezone.utc) - last_used["LastUsedDate"]).days
                if age > threshold_days:
                    roles.append(role["RoleName"])
        if roles_response["IsTruncated"]:
            roles_response = session_objects["iam_client"].list_roles(
                Marker=roles_response["Marker"]
            )
        else:
            break

    logger.info(f"{len(roles)} roles found")
    logger.debug(roles)

    return roles



import datetime


def preprocess_data(data):
    """Preprocess data for table output."""

    if isinstance(data, dict):
        data = [data]


def table_from_list_dict_data(data):
    """Print a table from a list of dictionaries."""

    headers = sorted(
        set(key for item in data for key in item.keys())
    )  # Sort headers for consistent ordering
    column_widths = {
        header: max(len(header), max(len(str(item.get(header, ""))) for item in data))
        for header in headers
    }

    # Print table header
    header_row = (
        "+" + "+".join(["-" * (column_widths[header] + 2) for header in headers]) + "+"
    )
    print(header_row)
    header_content = "|".join(
        [" " + header.ljust(column_widths[header]) + " " for header in headers]
    )
    print(f"|{header_content}|")
    print(header_row.replace("-", "-"))

    # Print each row of data without dividers between rows
    for item in data:
        row_data = [
            " " + str(item.get(header, "")).ljust(column_widths[header]) + " "
            for header in headers
        ]
        print(f"|{'|'.join(row_data)}|")


def table_from_list_data(data):
    """Print a table from a list of strings."""

    header = "Resources"
    column_width = max(len(header), max(len(item) for item in data))

    # Print table header
    header_row = f"+{'-' * (column_width + 2)}+"
    print(header_row)
    print(f"| {header.center(column_width)} |")
    print(header_row.replace("-", "-"))

    # Print each row without dividers between rows
    for item in data:
        print(f"| {item.center(column_width)} |")


def output_table(data):
    """Print a table for CLI output."""

    if isinstance(data, datetime.datetime):
        table_from_list_data([str(data)])
    elif isinstance(data, dict):
        table_from_list_dict_data([data])
    elif all(isinstance(item, dict) for item in data):
        table_from_list_dict_data(data)
    elif all(isinstance(item, str) for item in data):
        table_from_list_data(data)
    else:
        print("Mixed or unsupported data format.")



import logging

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate

logger = logging.getLogger(__name__)

action_description = "Create an authenticated session."

DEFAULTS = {
    "access_key": None,
    "secret_key": None,
    "session_token": None,
    "profile": None,
    "role_arn": None,
    "mfa_serial": None,
    "mfa_token": None,
    "debug": False,
    "silent": False,
}
RULES = [{"and": ["mfa_token", "mfa_serial"]}, {"and": ["access_key", "secret_key"]}]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def create_session(**kwargs):
    """Create an authenticated session"""
    return kwargs["session"]



import logging
from datetime import datetime, timezone

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {"region": None, "debug": False, "silent": False}
RULES = [{"required": ["threshold_days"]}]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def discover_old_access_keys(**kwargs):
    """Discover access keys over a certain age"""

    # Required parameters
    threshold_days = kwargs.pop("threshold_days")

    session_objects = set_session_objects(
        kwargs["session"], clients=["iam"], region=kwargs["region"]
    )

    logger.info(f"Discovering access keys over {threshold_days} days old")

    access_keys = []
    users_response = session_objects["iam_client"].list_users(MaxItems=50)
    while True:
        for user in users_response["Users"]:
            try:
                keys_response = session_objects["iam_client"].list_access_keys(
                    UserName=user["UserName"]
                )
            except Exception as e:
                logger.error(f"Failed to list access keys for {user['UserName']} - {e}")
                continue
            for key in keys_response["AccessKeyMetadata"]:
                if key["Status"] == "Active":
                    age = (datetime.now(timezone.utc) - key["CreateDate"]).days
                    if age > threshold_days:
                        access_keys.append(
                            {
                                "UserName": user["UserName"],
                                "AccessKeyId": key["AccessKeyId"],
                            }
                        )
        if users_response["IsTruncated"]:
            users_response = session_objects["iam_client"].list_users(
                Marker=users_response["Marker"]
            )
        else:
            break

    logger.info(f"{len(access_keys)} access keys found")
    logger.debug(access_keys)

    return access_keys



from avtomat_aws.helpers.cli.set_output import set_output
from avtomat_aws.services.ec2 import share_snapshots

ACTION_DESCRIPTION = "Share EC2 snapshots with other accounts."


def add_cli_arguments(parser):
    """Argument parsing"""

    required = parser.add_argument_group("required")

    required.add_argument(
        "--snapshot_ids",
        nargs="+",
        help="EC2 snapshot IDs to move, separated by space.",
        required=True,
    )
    required.add_argument(
        "--target_account", help="Target account for the snapshots.", required=True
    )


def cli(args):
    """Command-line interface function"""

    inputs = vars(args)

    try:
        result = share_snapshots(**inputs)
        set_output(result, inputs)
    except Exception as e:
        print(f"Action failed - {e}")
        exit(1)



import logging

from avtomat_aws.loggers.set_logging import set_logging


def set_logger():
    """Decorator to set a logger for the action."""

    def decorator(func):
        def wrapper(**kwargs):
            # Setup logger
            set_logging()
            logger = logging.getLogger("avtomat_aws")
            # Set log level
            if kwargs.get("debug"):
                logger.setLevel(logging.DEBUG)
            if kwargs.get("silent"):
                logger.setLevel(logging.WARNING)
            # Call the action
            return func(**kwargs)

        return wrapper

    return decorator



from avtomat_aws.helpers.cli.set_output import set_output
from avtomat_aws.services.ec2 import create_snapshots

ACTION_DESCRIPTION = "Create EBS snapshots from volumes."


def add_cli_arguments(parser):
    """Argument parsing"""

    required = parser.add_argument_group("required")

    required.add_argument(
        "--volume_ids",
        nargs="+",
        help="Volume IDs to create snapshots for, separated by space.",
        required=True,
    )


def cli(args):
    """Command-line interface function"""

    inputs = vars(args)

    try:
        result = create_snapshots(**inputs)
        set_output(result, inputs)
    except Exception as e:
        print(f"Action failed - {e}")
        exit(1)



import logging
import uuid

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {"reboot": False, "region": None, "debug": False, "silent": False}
RULES = [{"required": ["instance_ids"]}]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def create_images(**kwargs):
    """Create EC2 images (AMI)"""

    # Required parameters
    instance_ids = kwargs.pop("instance_ids")

    if not instance_ids:
        logger.info("No instances to create images from")
        return []

    session_objects = set_session_objects(
        kwargs["session"], resources=["ec2"], region=kwargs["region"]
    )

    logger.info(f"Creating images")

    images = []
    failed = []
    for instance_id in instance_ids:
        try:
            instance = session_objects["ec2_resource"].Instance(instance_id)
            # Generate random uuid as image names must be unique
            uuid_str = str(uuid.uuid4())
            image = instance.create_image(
                Name=f"source_{instance_id}_{uuid_str}",
                Description=f"Image of {instance_id}",
                NoReboot=not kwargs["reboot"],
                TagSpecifications=[
                    {
                        "ResourceType": "image",
                        "Tags": get_tags(instance_id, session_objects),
                    }
                ],
            )
            logger.info(f"{instance_id} - creating {image.id}")
            images.append(image.id)
        except Exception as e:
            failed.append(instance_id)
            logger.error(f"{instance_id} - failed to create image - {e}")
            continue

    # An error is returned after 10 minutes of waiting. Can be modified with the WaiterConfig.
    for image in images:
        image = session_objects["ec2_resource"].Image(image)
        image.wait_until_exists(Filters=[{"Name": "state", "Values": ["available"]}])

    logger.info(f"{len(images)} images created")
    if failed:
        logger.warning(f"{len(failed)} instances failed the image creation process")
        logger.debug(failed)

    return images


def get_tags(instance_id, session_objects):
    """Get tags from an EC2 instance"""
    instance = session_objects["ec2_resource"].Instance(instance_id)
    tags = instance.tags
    return tags



from datetime import datetime, timedelta


def required_rule(kwargs, params):
    """Validates that required parameters are provided."""

    for param in params:
        if param not in kwargs or kwargs[param] is None:
            raise ValueError(f"Required parameter '{param}' not specified")


def choice_rule(kwargs, params):
    """Validates that parameter values are within allowed choices."""

    for param_dict in params:
        for param, choices in param_dict.items():
            if kwargs[param] is not None and type(kwargs[param]) is list:
                for value in kwargs[param]:
                    if value not in choices:
                        raise ValueError(
                            f"Invalid value for '{param}' - '{value}'. Valid values: {choices}"
                        )
            else:
                if kwargs[param] is not None and kwargs[param] not in choices:
                    raise ValueError(
                        f"Invalid value for '{param}' - '{kwargs[param]}'. Valid values: {choices}"
                    )


def and_rule(kwargs, params):
    """Validates that if either 'x' or 'y' is provided, then both must be provided."""

    try:
        x, y = params
    except ValueError:
        raise ValueError("Rule 'and' requires two parameters")

    if (kwargs.get(x) and not kwargs.get(y)) or (not kwargs.get(x) and kwargs.get(y)):
        raise ValueError(f"'{x}' and '{y}' must be used together")


def at_most_one_rule(kwargs, params):
    """Validates that at most one of the parameters is provided."""

    if len(params) < 2:
        raise ValueError("Rule 'at_most_one' requires at least two parameters")

    params_defined = sum(kwargs.get(param) not in (None, [], False) for param in params)
    if params_defined > 1:
        raise ValueError(
            f"The following parameters cannot be used together, select only one: {params}"
        )


def at_least_one_rule(kwargs, params):
    """Validates that at least one of the parameters is provided."""

    for param in params:
        if kwargs.get(param):
            return
    raise ValueError(f"At least one of the following parameters is required: {params}")


def logging_rule(kwargs):
    """Validates that 'debug' and 'silent' log levels are not both set."""

    if kwargs.get("debug") and kwargs.get("silent"):
        raise ValueError("Cannot set both 'debug' and 'silent' log levels")


def date_yymmdd_rule(kwargs, params):
    """Validates that date parameters are in the correct format (YYYY/MM/DD)."""

    for param in params:
        if kwargs.get(param) and type(kwargs[param]) is str:
            try:
                datetime.strptime(kwargs[param], "%Y/%m/%d")
            except ValueError:
                raise ValueError(
                    f"'{param}' '{kwargs[param]}' does not match the format YYYY/MM/DD"
                )


def cloudtrail_rule(kwargs, params):
    """Validates that a date is not in the future or not more than 90 days in the past."""

    for param in params:
        if type(kwargs[param]) is str:
            param_date = datetime.strptime(kwargs[param], "%Y/%m/%d")
        else:
            param_date = kwargs[param]

        if param_date < datetime.now() - timedelta(days=90):
            raise ValueError(
                f"'{param}' '{kwargs[param]}' cannot be more than 90 days in the past"
            )
        if param_date > datetime.now():
            raise ValueError(f"'{param}' '{kwargs[param]}' cannot be in the future")



import logging

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {"region": None, "debug": False, "silent": False}
RULES = [{"required": ["security_group_ids"]}]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def delete_security_groups(**kwargs):
    """Delete EC2 security groups"""

    # Required parameters
    security_group_ids = kwargs.pop("security_group_ids")

    if not security_group_ids:
        logger.info("No security groups to delete")
        return []

    session_objects = set_session_objects(
        kwargs["session"], clients=["ec2"], region=kwargs["region"]
    )

    logger.info(f"Deleting security groups")

    failed = []
    counter = 0
    for security_group_id in security_group_ids:
        try:
            session_objects["ec2_client"].delete_security_group(
                GroupId=security_group_id
            )
            counter += 1
            logger.info(f"{security_group_id} - deleted")
        except Exception as e:
            failed.append(security_group_id)
            logger.error(f"{security_group_id} - failed to delete - {e}")

    logger.info(f"{counter} security groups deleted")
    if failed:
        logger.warning(f"{len(failed)} security groups failed to delete")
        logger.debug(failed)

    return failed



import logging

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {"region": None, "debug": False, "silent": False}
RULES = [{"required": ["volume_ids"]}]


@validate(DEFAULTS, RULES)
@set_logger()
@authenticate()
def create_snapshots(**kwargs):
    """Create EBS snapshots from volumes"""

    # Required parameters
    volume_ids = kwargs.pop("volume_ids")

    if not volume_ids:
        logger.info("No volumes to create snapshots from")
        return []

    session_objects = set_session_objects(
        kwargs["session"], resources=["ec2"], region=kwargs["region"]
    )

    logger.info(f"Creating snapshots")

    snapshots = []
    failed = []
    for volume_id in volume_ids:
        try:
            volume = session_objects["ec2_resource"].Volume(volume_id)
            snapshot = volume.create_snapshot(
                Description=f"Snapshot of {volume_id}",
                TagSpecifications=[
                    {
                        "ResourceType": "snapshot",
                        "Tags": [{"Key": "Name", "Value": f"source:{volume_id}"}],
                    }
                ],
            )
            logger.info(f"{volume_id} - creating {snapshot.id}")
            snapshots.append(snapshot.id)
        except Exception as e:
            failed.append(volume_id)
            logger.error(f"{volume_id} - failed to create snapshot - {e}")
            continue

    for snapshot in snapshots:
        snapshot = session_objects["ec2_resource"].Snapshot(snapshot)
        snapshot.wait_until_completed()

    logger.info(f"{len(snapshots)} snapshots created")
    if failed:
        logger.warning(f"{len(failed)} volumes failed the snapshot creation process")
        logger.debug(failed)

    return snapshots



from .rules import *

config = {
    "required": required_rule,
    "choice": choice_rule,
    "and": and_rule,
    "at_most_one": at_most_one_rule,
    "at_least_one": at_least_one_rule,
    "logging": logging_rule,
    "date_yymmdd": date_yymmdd_rule,
    "cloudtrail": cloudtrail_rule,
}



from avtomat_aws.helpers.cli.set_output import set_output
from avtomat_aws.services.ec2 import discover_volumes

ACTION_DESCRIPTION = "Discover EBS volumes."


def add_cli_arguments(parser):
    """Argument parsing"""

    parser.add_argument(
        "--instance_ids",
        nargs="+",
        help="Get volumes attached to specified instances.",
        required=False,
    )
    parser.add_argument(
        "--volume_ids",
        nargs="+",
        help="Volume IDs to focus on, separated by space.",
        required=False,
    )
    parser.add_argument(
        "--unencrypted",
        action="store_true",
        help="Get only unencrypted volumes.",
        required=False,
    )
    parser.add_argument(
        "--detached",
        action="store_true",
        help="Get only detached volumes.",
        required=False,
    )
    parser.add_argument(
        "--types",
        nargs="+",
        help="Volume types to check, separated by space.",
        required=False,
    )
    parser.add_argument(
        "--root",
        action="store_true",
        help="Get only root volumes.",
        required=False,
    )


def cli(args):
    """Command-line interface function"""

    inputs = vars(args)

    try:
        result = discover_volumes(**inputs)
        set_output(result, inputs)
    except Exception as e:
        print(f"Action failed - {e}")
        exit(1)



import logging

from avtomat_aws.decorators.authenticate import authenticate
from avtomat_aws.decorators.set_logger import set_logger
from avtomat_aws.decorators.validate import validate
from avtomat_aws.helpers.set_session_objects import set_session_objects

logger = logging.getLogger(__name__)

DEFAULTS = {"region": None, "debug": False, "silent": False}


@validate(DEFAULTS)
@set_logger()
@authenticate()
def discover_default_ebs_encryption(**kwargs):
    """Discover the default EBS encryption setting"""

    session_objects = set_session_objects(
        kwargs["session"], clients=["ec2"], region=kwargs["region"]
    )

    logger.info("Discovering default EBS encryption")

    enabled = session_objects["ec2_client"].get_ebs_encryption_by_default()[
        "EbsEncryptionByDefault"
    ]
    if enabled:
        kms_key_id = session_objects["ec2_client"].get_ebs_default_kms_key_id()[
            "KmsKeyId"
        ]
        logger.info(f"{kwargs['region']}: Default EBS encryption is enabled")
        logger.info(f"{kwargs['region']}: Default KMS key: {kms_key_id}")

        return {"Region": kwargs["region"], "Enabled": enabled, "KmsKeyId": kms_key_id}

    else:
        logger.info(f"{kwargs['region']}: Default EBS encryption is disabled")

        return {"Region": kwargs["region"], "Enabled": enabled, "KmsKeyId": None}
